open Prolog

(** Conjunction of goals to be proved.
  * This is our former type sk. *)
type goals = goal list

(** Disjunction of alternative goals, with state that should be restored
  * before backtracking.
  * This is more or less our former type fk except that we have compacted
  * potentially many restore steps into exactly one. This comes with a
  * slight change in the save/restore strategy. *)
type alts = (Term.state * goals) list


let search program goals alts =

  (** The function apply_fk has been inlined and apply_sk has been
    * merged with search which directly takes a list of goals. *)
  let rec search goals alts = match goals with
    | Atom (p,l) :: goals ->
        search (lookup program p l :: goals) alts
    | Or (a,b) :: goals ->
        search (a :: goals) ((Term.save (), b :: goals) :: alts)
    | And (a,b) :: goals ->
        search (a :: b :: goals) alts
    | Equals (u,v) :: goals ->
        let unifies =
          try Unify.unify u v ; true with
            | Unify.Unification_failure -> false
        in
          if unifies then
            search goals alts
          else begin
            match alts with
              | [] -> false
              | (s,goals)::alts -> Term.restore s ; search goals alts
          end
    | False :: _ ->
        begin match alts with
          | [] -> false
          | (s,goals)::alts -> Term.restore s ; search goals alts
        end
    | True :: goals -> search goals alts
    | [] -> true
  in
    search goals alts

let has_solution program goal =
  let s = Term.save () in
  let r = search program [goal] [s,[False]] in
    if r then Term.restore s ;
    r
