(** Les types ci-dessous sont simplifiés au maximum. Il était possible
  * d'avoir des types (concrets) distincts pour in_term et term, mais
  * j'ai choisi ici de les identifier, tout en les distinguant dans
  * l'interface en rendant in_term abstrait.
  * L'invariant imposé est que pour un term (vu de l'extérieur) Var
  * n'est utilisé que pour une variable non instantiée
  * (une vraie variable) tandis que cela n'est bien sûr pas forcé
  * pour in_term. *)

type term =
  | Fun of string * term list
  | Var of link ref
and link = term option (* None ssi c'est une variable non instantiée *)

type var = link ref

type in_term = term

(** Gestion de l'historique *)

type state = int
let state : (link ref) Stack.t = Stack.create ()
let stack_len = ref 0

let save () = !stack_len

let restore n =
  for i = 1 to !stack_len - n do
    let r = Stack.pop state in
      r := None
  done ;
  stack_len := n

(** Instantiation et déréférencement *)

let rec deref t = match t with
  | Var { contents = Some t } -> deref t
  | _ -> t

let bind x t =
  assert (!x = None) ;
  Stack.push x state ;
  incr stack_len ;
  x := Some (deref t)

let rec observe = deref

(** Constructeurs *)

let make f l = Fun (f,l)
let fresh () = ref None
let var v = Var v

(** Egalités *)

let var_equals v w = v == w

let rec equals t1 t2 = match observe t1, observe t2 with
  | Var v1, Var v2 -> v1 == v2
  | Fun (f1,l1), Fun (f2,l2) when f1 = f2 -> List.for_all2 equals l1 l2
  | _ -> false

(** Pretty printing *)

let rec pp_list pp sep chan = function
  | [] -> ()
  | [x] -> pp chan x
  | x::tl ->
      pp chan x ;
      Format.pp_print_string chan sep ;
      pp_list pp sep chan tl

let rec pp chan = function
  | Var { contents = Some t } ->
      pp chan t
  | Var { contents = None } ->
      Format.pp_print_string chan "<var>"
  | Fun (f,l) ->
      Format.pp_print_string chan f ;
      Format.pp_print_string chan "(" ;
      pp_list pp "," chan l ;
      Format.pp_print_string chan ")"
